@@ -6,9 +6,10 @@ import logging
 
 from PyQt5.QtCore import QTimer, QSettings, QModelIndex, Qt, QCoreApplication, QObject, pyqtSignal
 from PyQt5.QtGui import QIcon, QFont
-from PyQt5.QtWidgets import QMainWindow, QApplication, QFileDialog, QMessageBox, QStyledItemDelegate, QMenu
+from PyQt5.QtWidgets import QMainWindow, QApplication, QFileDialog, QMessageBox, QStyledItemDelegate, QMenu, QAction
 
-from opcua import ua
+
+from asyncua import ua
 
 from uawidgets import resources
 from uawidgets.attrs_widget import AttrsWidget
@@ -42,12 +43,13 @@ class BoldDelegate(QStyledItemDelegate):
         QStyledItemDelegate.paint(self, painter, option, idx)
 
 
-class ActionsManager(object):
+class ActionsManager:
     """
     Manage actions of Modeler
     """
 
-    def __init__(self, ui, model_mgr):
+    def __init__(self, window, ui, model_mgr):
+        self.window = window
         self.ui = ui
         self.model_mgr = model_mgr
 
@@ -56,6 +58,7 @@ class ActionsManager(object):
         self.ui.actionNew.triggered.connect(self.model_mgr.new)
         self.ui.actionOpen.triggered.connect(self.model_mgr.open)
         self.ui.actionCopy.triggered.connect(self.model_mgr.copy)
+        self.ui.actionQuit.triggered.connect(self.window.close)
         self.ui.actionPaste.triggered.connect(self.model_mgr.paste)
         self.ui.actionDelete.triggered.connect(self.model_mgr.delete)
         self.ui.actionImport.triggered.connect(self.model_mgr.import_xml)
@@ -93,17 +96,17 @@ class ActionsManager(object):
 
     def update_actions_states(self, node):
         self.disable_add_actions()
-        if not node or node in (self.model_mgr.get_current_server().nodes.root, 
-                                self.model_mgr.get_current_server().nodes.types, 
-                                self.model_mgr.get_current_server().nodes.event_types, 
-                                self.model_mgr.get_current_server().nodes.object_types, 
-                                self.model_mgr.get_current_server().nodes.reference_types, 
-                                self.model_mgr.get_current_server().nodes.variable_types, 
+        if not node or node in (self.model_mgr.get_current_server().nodes.root,
+                                self.model_mgr.get_current_server().nodes.types,
+                                self.model_mgr.get_current_server().nodes.event_types,
+                                self.model_mgr.get_current_server().nodes.object_types,
+                                self.model_mgr.get_current_server().nodes.reference_types,
+                                self.model_mgr.get_current_server().nodes.variable_types,
                                 self.model_mgr.get_current_server().nodes.data_types):
             return
         path = node.get_path()
-        nodeclass = node.get_node_class()
-        typedefinition = node.get_type_definition()
+        nodeclass = node.read_node_class()
+        typedefinition = node.read_type_definition()
 
         self.ui.actionCopy.setEnabled(True)
         self.ui.actionDelete.setEnabled(True)
@@ -128,6 +131,8 @@ class ActionsManager(object):
             self.ui.actionAddDataType.setEnabled(True)
             if self.model_mgr.get_current_server().nodes.enum_data_type in path:
                 self.ui.actionAddProperty.setEnabled(True)
+            elif self.model_mgr.get_current_server().nodes.base_structure_type in path:
+                self.ui.actionAddVariable.setEnabled(True)
             return  # not other nodes should be added here
 
         self.ui.actionAddFolder.setEnabled(True)
@@ -242,13 +247,17 @@ class ModelManagerUI(QObject):
             self._last_model_dir = os.path.dirname(path)
             self.settings.setValue("last_model_dir", self._last_model_dir)
         self._model_mgr.open(path)
+        self.modeler.update_recent_files(path)
+
+    def open_file(self, path):
+        self._model_mgr.open(path)
 
     @trycatchslot
     def import_xml(self):
         last_import_dir = self.settings.value("last_import_dir", ".")
         path, ok = QFileDialog.getOpenFileName(self.modeler, caption="Import reference OPC UA XML", filter="XML Files (*.xml *.XML)", directory=last_import_dir)
         if not ok:
-            return None
+            return
         self.settings.setValue("last_import_dir", last_import_dir)
         self._model_mgr.import_xml(path)
 
@@ -263,7 +272,8 @@ class ModelManagerUI(QObject):
                 self._last_model_dir = os.path.dirname(path)
                 self.settings.setValue("last_model_dir", self._last_model_dir)
             self._model_mgr.save_xml(path)
-            self._model_mgr.save_ua_model(path)
+            path = self._model_mgr.save_ua_model(path)
+            self.modeler.update_recent_files(path)
 
     @trycatchslot
     def save(self):
@@ -277,7 +287,8 @@ class ModelManagerUI(QObject):
     def add_method(self):
         args, ok = NewUaMethodDialog.getArgs(self.modeler, "Add Method", self._model_mgr.server_mgr)
         if ok:
-            self._model_mgr.add_method(*args)
+            nodes = self._model_mgr.add_method(*args)
+            self._add_modelling_rule(nodes)
 
     @trycatchslot
     def add_object_type(self):
@@ -289,34 +300,45 @@ class ModelManagerUI(QObject):
     def add_folder(self):
         args, ok = NewNodeBaseDialog.getArgs(self.modeler, "Add Folder", self._model_mgr.server_mgr)
         if ok:
-            self._model_mgr.add_folder(*args)
+            node = self._model_mgr.add_folder(*args)
+            self._add_modelling_rule(node)
 
     @trycatchslot
     def add_object(self):
         args, ok = NewUaObjectDialog.getArgs(self.modeler, "Add Object", self._model_mgr.server_mgr, base_node_type=self._model_mgr.server_mgr.nodes.base_object_type)
         if ok:
-            self._model_mgr.add_object(*args)
+            nodes = self._model_mgr.add_object(*args)
+            # FIXME: in this particular case we may want to navigate recursively to add ref
+            self._add_modelling_rule(nodes)
+
+    def _add_modelling_rule(self, nodes):
+        if not isinstance(nodes, (list, tuple)):
+            nodes = [nodes]
+        for node in nodes:
+            path = node.get_path()
+            if self._model_mgr.server_mgr.nodes.base_object_type in path:
+                # we are creating a new type, add modeling rule
+                node.set_modelling_rule(True)
 
     @trycatchslot
     def add_data_type(self):
         args, ok = NewNodeBaseDialog.getArgs(self.modeler, "Add Data Type", self._model_mgr.server_mgr)
         if ok:
             self._model_mgr.add_data_type(*args)
-    
+
     @trycatchslot
     def add_variable(self):
-        dtype = self.settings.value("last_datatype", None)
-        args, ok = NewUaVariableDialog.getArgs(self.modeler, "Add Variable", self._model_mgr.server_mgr, default_value=9.99, dtype=dtype)
+        args, ok = NewUaVariableDialog.getArgs(self.modeler, "Add Variable", self._model_mgr.server_mgr)
         if ok:
-            self._model_mgr.add_variable(*args)
-            self.settings.setValue("last_datatype", args[4])
+            node = self._model_mgr.add_variable(*args)
+            self._add_modelling_rule(node)
 
     @trycatchslot
     def add_property(self):
-        dtype = self.settings.value("last_datatype", None)
-        args, ok = NewUaVariableDialog.getArgs(self.modeler, "Add Property", self._model_mgr.server_mgr, default_value=9.99, dtype=dtype)
+        args, ok = NewUaVariableDialog.getArgs(self.modeler, "Add Property", self._model_mgr.server_mgr)
         if ok:
-            self._model_mgr.add_property(*args)
+            node = self._model_mgr.add_property(*args)
+            self._add_modelling_rule(node)
 
     @trycatchslot
     def add_variable_type(self):
@@ -328,7 +350,7 @@ class ModelManagerUI(QObject):
 class UaModeler(QMainWindow):
     """
     Main class of modeler. Should be as simple as possible, try to push things to other classes
-    or even better python-opcua
+    or even better python-asyncua
     """
 
     def __init__(self):
@@ -346,7 +368,7 @@ class UaModeler(QMainWindow):
         self.settings = QSettings()
 
         self._restore_ui_geometri()
-        
+
         self.tree_ui = TreeWidget(self.ui.treeView)
         self.tree_ui.error.connect(self.show_error)
 
@@ -369,7 +391,7 @@ class UaModeler(QMainWindow):
         self.model_mgr = ModelManagerUI(self)
         self.model_mgr.error.connect(self.show_error)
         self.model_mgr.titleChanged.connect(self.update_title)
-        self.actions = ActionsManager(self.ui, self.model_mgr)
+        self.actions = ActionsManager(self, self.ui, self.model_mgr)
 
         self.setup_context_menu_tree()
 
@@ -377,6 +399,39 @@ class UaModeler(QMainWindow):
         self.ui.treeView.setItemDelegate(delegate)
         self.ui.treeView.selectionModel().currentChanged.connect(self._update_actions_state)
 
+        self._recent_files = self.settings.value("recent_files", [])
+        self._recent_files_max_count = int(self.settings.value("recent_files_max_count", 10))
+        self._recent_files_acts = [QAction(self, visible=False, triggered=self.open_recent_files) for _ in range(self._recent_files_max_count)]
+        for act in self._recent_files_acts:
+            self.ui.menuRecentFiles.addAction(act)
+        self._update_recent_files_ui()
+
+    def open_recent_files(self):
+        if not self.model_mgr.try_close_model():
+            return
+        action = self.sender()
+        if action:
+            path = action.data()
+            self.model_mgr.open_file(path)
+            self.update_recent_files(path)
+
+    def update_recent_files(self, path):
+        if self._recent_files and path == self._recent_files[0]:
+            return
+        if self._recent_files is not None:
+            if path in self._recent_files:
+                self._recent_files.remove(path)
+            self._recent_files.insert(0, path)
+            self._recent_files = self._recent_files[:self._recent_files_max_count]
+            self._update_recent_files_ui()
+
+    def _update_recent_files_ui(self):
+        if self._recent_files is not None:
+            for idx, path in enumerate(self._recent_files):
+                self._recent_files_acts[idx].setText(path)
+                self._recent_files_acts[idx].setData(path)
+                self._recent_files_acts[idx].setVisible(True)
+
     def get_current_node(self, idx=None):
         return self.tree_ui.get_current_node(idx)
 
@@ -420,7 +475,6 @@ class UaModeler(QMainWindow):
         self._contextMenu.addAction(self.ui.actionAddDataType)
 
     def _show_context_menu_tree(self, position):
-        print("SHOW REQUEST")
         node = self.tree_ui.get_current_node()
         if node:
             self._contextMenu.exec_(self.ui.treeView.viewport().mapToGlobal(position))
@@ -452,14 +506,16 @@ class UaModeler(QMainWindow):
     @trycatchslot
     def show_refs(self, idx=None):
         node = self.get_current_node(idx)
-        self.refs_ui.show_refs(node)
+        if node:
+            self.refs_ui.show_refs(node)
 
     @trycatchslot
     def show_attrs(self, idx=None):
         if not isinstance(idx, QModelIndex):
             idx = None
         node = self.get_current_node(idx)
-        self.attrs_ui.show_attrs(node)
+        if node:
+            self.attrs_ui.show_attrs(node)
 
     def nodesets_change(self, data):
         self.idx_ui.reload()
@@ -481,6 +537,7 @@ class UaModeler(QMainWindow):
         self.settings.setValue("splitter_left", self.ui.splitterLeft.saveState())
         self.settings.setValue("splitter_right", self.ui.splitterRight.saveState())
         self.settings.setValue("splitter_center", self.ui.splitterCenter.saveState())
+        self.settings.setValue("recent_files", self._recent_files)
         event.accept()
 
 
@@ -491,7 +548,7 @@ def main():
     logging.getLogger().addHandler(handler)
     logging.getLogger("uamodeler").setLevel(logging.INFO)
     logging.getLogger("uawidgets").setLevel(logging.INFO)
-    #logging.getLogger("opcua").setLevel(logging.INFO)  # to enable logging of ua server
+    #logging.getLogger("asyncua").setLevel(logging.INFO)  # to enable logging of ua server
     modeler.show()
     sys.exit(app.exec_())
 
